/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "lib.h"
#include "lock.h"
#include "auth.h"

/* Encryption and Decryption of the challenge key */
#ifdef HAVE_SSL

#ifndef __APPLE_CC__
#include <openssl/md5.h>
#include <openssl/blowfish.h>
#include <openssl/rand.h>
#else                           /* YAY - We're MAC OS */
#include <sys/md5.h>
#include <crypto/blowfish.h>
#endif                          /* __APPLE_CC__ */

void gen_chal(char *buf)
{
   RAND_bytes(buf, VTUN_CHAL_SIZE);
}

void encrypt_chal(char *chal, char *pwd)
{
   register int i;
   BF_KEY key;

   BF_set_key(&key, 16, MD5(pwd, strlen(pwd), NULL));

   for (i = 0; i < VTUN_CHAL_SIZE; i += 8)
      BF_ecb_encrypt(chal + i, chal + i, &key, BF_ENCRYPT);
}

void decrypt_chal(char *chal, char *pwd)
{
   register int i;
   BF_KEY key;

   BF_set_key(&key, 16, MD5(pwd, strlen(pwd), NULL));

   for (i = 0; i < VTUN_CHAL_SIZE; i += 8)
      BF_ecb_encrypt(chal + i, chal + i, &key, BF_DECRYPT);
}

#else                           /* HAVE_SSL */

void encrypt_chal(char *chal, char *pwd)
{
   char *xor_msk = pwd;
   register int i, xor_len = strlen(xor_msk);

   for (i = 0; i < VTUN_CHAL_SIZE; i++)
      chal[i] ^= xor_msk[i % xor_len];
}

void inline decrypt_chal(char *chal, char *pwd)
{
   encrypt_chal(chal, pwd);
}

/* Generate PSEUDO random challenge key. */
void gen_chal(char *buf)
{
   register int i;

   srand(time(NULL));

   for (i = 0; i < VTUN_CHAL_SIZE; i++)
      buf[i] = (unsigned int) (255.0 * rand() / RAND_MAX);
}
#endif                          /* HAVE_SSL */

/* 
 * Functions to convert binary flags to character string.
 * string format:  <CS64> 
 * C - compression, S - speed for shaper and so on.
 */

char *bf2cf(struct vtun_host *host)
{
   static char str[50], *ptr = str;

   *(ptr++) = '<';

   switch (host->flags & VTUN_PROT_MASK) {
   case VTUN_TCP:
      *(ptr++) = 'T';
      break;

   case VTUN_UDP:
      *(ptr++) = 'U';
      break;
   }

   switch (host->flags & VTUN_TYPE_MASK) {
   case VTUN_TTY:
      *(ptr++) = 't';
      break;

   case VTUN_PIPE:
      *(ptr++) = 'p';
      break;

   case VTUN_ETHER:
      *(ptr++) = 'e';
      break;

   case VTUN_TUN:
      *(ptr++) = 'u';
      break;
   }

   if ((host->flags & VTUN_PPS) /* && host->pps_in */ )
      ptr += sprintf(ptr, "P%d", host->pps_in);
   if ((host->flags & VTUN_SHAPE) /* && host->spd_in */ )
      ptr += sprintf(ptr, "S%d", host->spd_in);

   if (host->flags & VTUN_ZLIB)
      ptr += sprintf(ptr, "C%d", host->zlevel);

   if (host->flags & VTUN_LZO)
      ptr += sprintf(ptr, "L%d", host->zlevel);

   if (host->flags & VTUN_KEEP_ALIVE)
      *(ptr++) = 'K';

   if (host->flags & VTUN_ENCRYPT)
      *(ptr++) = 'E';

   if (host->flags & VTUN_PERSIST)
      *(ptr++) = 'f';
   *(ptr++) = 'M';
   *(ptr++) = host->i_max + '0';

   strcat(ptr++, ">");

   sprintf(ptr, "( %d %d %d %d %s )", host->ka_interval, host->ka_failure, host->ka_uptime, host->ka_checktime, host->mode);

   return str;
}

/* return 1 on success, otherwise 0 */

int cf2bf(char *str, struct vtun_host *host)
{
   char *ptr, *p;
   int s;

   if ((ptr = strchr(str, '<'))) {
      ptr++;
      while (*ptr) {
         switch (*ptr++) {
         case 'M':
            {
               char *p = (ptr++);
               int i = p[0] - '0';
               if (i > 0 && i < vtun.svr_num)
                  vtun.svr_num = i;
            }
            break;
         case 'f':
            host->flags |= VTUN_PERSIST;
            break;
         case 't':
            host->flags |= VTUN_TTY;
            break;
         case 'p':
            host->flags |= VTUN_PIPE;
            break;
         case 'e':
            host->flags |= VTUN_ETHER;
            break;
         case 'u':
            host->flags |= VTUN_TUN;
            break;
         case 'U':
            host->flags &= ~VTUN_PROT_MASK;
            host->flags |= VTUN_UDP;
            break;
         case 'T':
            host->flags &= ~VTUN_PROT_MASK;
            host->flags |= VTUN_TCP;
            break;
         case 'K':
            host->flags |= VTUN_KEEP_ALIVE;
            break;
         case 'C':
            if ((s = strtol(ptr, &p, 10)) == ERANGE || ptr == p)
               return 0;
            host->flags |= VTUN_ZLIB;
            host->zlevel = s;
            ptr = p;
            break;
         case 'L':
            if ((s = strtol(ptr, &p, 10)) == ERANGE || ptr == p)
               return 0;
            host->flags |= VTUN_LZO;
            host->zlevel = s;
            ptr = p;
            break;
         case 'E':
            host->flags |= VTUN_ENCRYPT;
            break;
         case 'S':
            if ((s = strtol(ptr, &p, 10)) == ERANGE || ptr == p)
               return 0;
            if (s) {
               host->flags |= VTUN_SHAPE;
               host->spd_out = s;
            }
            ptr = p;
            break;
         case 'P':
            if ((s = strtol(ptr, &p, 10)) == ERANGE || ptr == p)
               return 0;
            if (s) {
               host->flags |= VTUN_PPS;
               host->pps_out = s;
            }
            ptr = p;
            break;
         case '>':
            {
               //vtun_syslog(LOG_DEBUG, "%s", ptr);
               char buf_mode[100];
               sscanf(ptr, "( %d %d %d %d %s )", &(host->ka_interval), &(host->ka_failure), &(host->ka_uptime),
                      &(host->ka_checktime), buf_mode);
               host->mode = strdup(buf_mode);
               //vtun_syslog(LOG_DEBUG, "(%d %d %d %d)", host->ka_interval, host->ka_failure, host->ka_uptime, host->ka_checktime);
            }
            return 1;
         default:
            return 0;
         }
      }
   }
   return 0;
}

/* 
 * Functions to convert binary key data to character string.
 * string format:  <char_data> 
 */

char *cl2cs(char *chal)
{
   static char str[VTUN_CHAL_SIZE * 2 + 3], *chr = "abcdefghijklmnop";
   register char *ptr = str;
   register int i;

   *(ptr++) = '<';
   for (i = 0; i < VTUN_CHAL_SIZE; i++) {
      *(ptr++) = chr[((chal[i] & 0xf0) >> 4)];
      *(ptr++) = chr[(chal[i] & 0x0f)];
   }

   *(ptr++) = '>';
   *ptr = '\0';

   return str;
}

int cs2cl(char *str, char *chal)
{
   register char *ptr = str;
   register int i;

   if (!(ptr = strchr(str, '<')))
      return 0;
   ptr++;
   if (!strtok(ptr, ">") || strlen(ptr) != VTUN_CHAL_SIZE * 2)
      return 0;

   for (i = 0; i < VTUN_CHAL_SIZE && *ptr; i++, ptr += 2) {
      chal[i] = (*ptr - 'a') << 4;
      chal[i] |= *(ptr + 1) - 'a';
   }

   return 1;
}
